"""
    deblaze - A remote method enumeration tool for flex servers
    Copyright (C) 2009 Jon Rose
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from twisted.internet import defer, reactor
import threading
from twisted.python import threadable
from twisted.protocols import portforward
import getopt, sys
sys.path.append("pyamf/")
import pyamf
from pyamf import remoting
import re
import logging
from outputManager import outputManager

#required for using threads with the Reactor
threadable.init()
om = outputManager()


class ReactorException(Exception):
      def __init__(self, *args):
            Exception.__init__(self, *args)
            

class ReactorThread(threading.Thread):    
      """Run the Reactor in a Thread to prevent blocking the 
         Main Thread once reactor.run is called"""
    
      def __init__(self, localport, desthost, destport):
              threading.Thread.__init__(self)
              self._reactorRunning = False
              self._localport = int(localport)
              self._desthost = desthost
              self._destport = int(destport)
              
      def run(self):
              if self._reactorRunning:
                    raise ReactorException("Reactor Already Running")
              
              self._reactorRunning = True
              
              #call run passing a False flag indicating to the
              #reactor not to install sig handlers since sig handlers
              #only work on the main thread

              reactor.listenTCP(self._localport, portforward.ProxyFactory(self._desthost, self._destport))
              reactor.run(False)              
        
      def callInReactor(self, callable, *args, **kw):
            if self._reactorRunning:
                  reactor.callFromThread(callable, *args, **kw)

            else:
                  callable(*args, **kw)                 
            
      def isReactorRunning(self):
              return self._reactorRunning
       
      def startReactor(self):
             if self._reactorRunning:
                    raise ReactorException("Reactor Already Running")
             
             threading.Thread.start(self)
             
             reactor.addSystemEventTrigger('after', 'shutdown', self.__reactorShutDown)
              
            
      def stopReactor(self):
            """may want a way to force thread to join if reactor does not shutdown
               properly"""
            
            if not self._reactorRunning:
                   raise ReactorException("Reactor Not Running")
           
            reactor.callFromThread(reactor.stop)
             
 
      def addReactorEventTrigger(self, phase, eventType, callable):
             if self._reactorRunning:
                reactor.callFromThread(reactor.addSystemEventTrigger, phase, eventType, callable)

             else:
                reactor.addSystemEventTrigger(phase, eventType, callable)
                    
                     
      def __reactorShuttingDown(self):
            pass

      def __reactorShutDown(self):
            """This method called when the reactor is stopped"""
            self._reactorRunning = False
            
      def startProtocol(self):
            self.transport.connect(desthost, destport)
      
      def startFuzz(self):
          print "Fuzz Signaled"      
            

#===============================================================================
# TCP proxy stuff
#===============================================================================
    
      def server_dataReceived(self, data):
    
            logging.info("Client ------> Server")
            #check for AMFPOST
            m = re.findall('^POST (.*) HTTP/.*Content-type: application/x-amf',data,re.I | re.S)
            if len(m) > 0:
                om.addGateway(str(m))
                payload = data.split('\r\n\r\n')[1]
                amfEnv = remoting.decode(payload)
                logging.info(amfEnv)
                
                for name, message in amfEnv:
                    if message.target == 'null':
                        logging.info("SWF generated request")
                        logging.info("Gateway: " + str(m[0]))
                        om.addGateway(str(m[0]))
                        logging.info( "Service: " + str(message.body[0].destination))
                        om.addService(message.body[0].destination) 
                        logging.info( "Method: " + str(message.body[0].operation))
                        om.addMethod(str(message.body[0].operation)) 
                        logging.info( "Params: " + str(message.body[0].body))
                    else:
                        logging.info("PYAMF request")
                        logging.info("Gateway: " + str(m[0]))
                        om.addGateway(str(m[0]))
                        logging.info("Service: " + message.target.split('.')[0])
                        om.addService(message.target.split('.')[0])
                        logging.info("Method: " + message.target.split('.')[1])
                        om.addMethod(message.target.split('.')[1]) 
                        logging.info("Params: " + str(message.body))
        
       # x = remotingObj(str(m[0]),message.body[0].destination,message.body[0].operation,paramlist);
       # print x.gateway
       # print x.service
       # print x.method
       # print x.params
                
                
            logging.debug("%r" % data)
    
            portforward.Proxy.dataReceived(self, data)
    
      portforward.ProxyServer.dataReceived = server_dataReceived
    
    
      def client_dataReceived(self, data):

            logging.info( "Server ------> Client")
            #check for AMF Response
            # BUG buffer incoming data for cases where AMF response is split between multiple requests
            m = re.findall('^HTTP.* 200 .*Content-Type: application/x-amf',data,re.DOTALL)
            if len(m) > 0:
                payload = data.split('\r\n\r\n')[1]
                amfmessage = remoting.decode(payload)
                logging.info( amfmessage)
    
            logging.debug("%r" % data)
        
            portforward.Proxy.dataReceived(self, data)
    
      portforward.ProxyClient.dataReceived = client_dataReceived
