"""
    deblaze - A remote method enumeration tool for flex servers
    Copyright (C) 2009 Jon Rose
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os, subprocess
import re
import urllib
import sys
from outputManager import outputManager

om = outputManager()

class swfdecompiler:
    
    def __init__(self):
        self.url= ''
        self.gatewaysArray = []
        self.servicesArray = []
        self.methodsArray = []
        self.commonfunctions = ['activate',
'addChild',
'addEventListener',
'addHandler',
'addStyleToProtoChain',
'appendChild',
'apply',
'attribute',
'attributes',
'callLater',
'charAt',
'child',
'children',
'concat',
'contains',
'copy',
'createAutomationIDPart',
'createCursor',
'dataForFormat',
'deactivate',
'deepClone',
'descendants',
'disconnect',
'dispatchEvent',
'elements',
'end',
'execute',
'getAssociatedFactory',
'getAutomationChildAt',
'getChildAt',
'getDate',
'getDefinitionByName',
'getFullYear',
'getItemAt',
'getModule',
'getMonth',
'getSandboxRoot',
'getStyle',
'hasComplexContent',
'indexOf',
'initThemeColor',
'internalSetCredentials',
'itemToLabel',
'length',
'load',
'localName',
'localToGlobal',
'logout',
'measure',
'move',
'name',
'nodeKind',
'notification',
'notifyStyleChangeInChildren',
'onTweenEnd',
'onTweenUpdate',
'parent',
'pop',
'push',
'recordAutomatableDragCancel',
'recordAutomatableDragDrop',
'recordAutomatableDragStart',
'regenerateStyleCache',
'remove',
'removeChild',
'removeEventListener',
'replayAutomatableEvent',
'resolveAutomationIDPart',
'setActualSize',
'setCredentials',
'setNotification',
'setStyle',
'setTweenHandlers',
'setVisible',
'setWidth',
'shift',
'slice',
'splice',
'split',
'styleChanged',
'stylesInitialized',
'substr',
'text',
'toFixed',
'toLowerCase',
'toString',
'toXMLString',
'updateDisplayList',
'updateParent',
'watcherFired',
'willTrigger',]



############################
# Returns unique list      #
############################
    def unique(self,s):
         """Return a list of the elements in s, but without duplicates.
    
         For example, unique([1,2,3,1,2,3]) is some permutation of [1,2,3],
         unique("abcabc") some permutation of ["a", "b", "c"], and
         unique(([1, 2], [2, 3], [1, 2])) some permutation of
         [[2, 3], [1, 2]].
    
         For best speed, all sequence elements should be hashable.  Then
         unique() will usually work in linear time.
    
         If not possible, the sequence elements should enjoy a total
         ordering, and if list(s).sort() doesn't raise TypeError it's
         assumed that they do enjoy a total ordering.  Then unique() will
         usually work in O(N*log2(N)) time.
    
         If that's not possible either, the sequence elements must support
         equality-testing.  Then unique() will usually work in quadratic
         time.
         """
    
         n = len(s)
         if n == 0:
             return []
    
         # Try using a dict first, as that's the fastest and will usually
         # work.  If it doesn't work, it will usually fail quickly, so it
         # usually doesn't cost much to *try* it.  It requires that all the
         # sequence elements be hashable, and support equality comparison.
         u = {}
         try:
             for x in s:
                 u[x] = 1
         except TypeError:
             del u  # move on to the next method
         else:
             return u.keys()
    
         # We can't hash all the elements.  Second fastest is to sort,
         # which brings the equal elements together; then duplicates are
         # easy to weed out in a single pass.
         # NOTE:  Python's list.sort() was designed to be efficient in the
         # presence of many duplicate elements.  This isn't true of all
         # sort functions in all languages or libraries, so this approach
         # is more effective in Python than it may be elsewhere.
         try:
             t = list(s)
             t.sort()
         except TypeError:
             del t  # move on to the next method
         else:
             assert n > 0
             last = t[0]
             lasti = i = 1
             while i < n:
                 if t[i] != last:
                     t[lasti] = last = t[i]
                     lasti += 1
                 i += 1
             return t[:lasti]
    
         # Brute force is all that's left.
         u = []
         for x in s:
             if x not in u:
                 u.append(x)
         return u
    
    
        
    ############################
    #   Extracts remoting methods from SWF    #
    ############################
    
    def findRemotingMethods(self,swf):
        if not os.path.isfile('swfdump'):
            print "Missing swfdump in current path, please copy for automated mode"
            sys.exit(1)
        
        swfsplit = swf.split('/')
        self.url = swfsplit[0] + '//' + swfsplit[2]
        om.out( "Downloading SWF for analysis: " + swf)
            
        webFile = urllib.urlopen(swf)
        localFile = open(swfsplit[-1], 'w')
        localFile.write(webFile.read())
        webFile.close()
        localFile.close()
    
        proc = subprocess.Popen(['./swfdump', '-D',swfsplit[-1],], 
                        shell=False, 
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        )
                        
        stdout_value = proc.communicate()[0]
    
        m = re.findall("flex.messaging.io",stdout_value)
        if len(m) is 0:
            print "No Remoting Libs found!"
            os._exit(1)
    
    
        m = re.findall("endpoint uri=\"http://{server.name}:{server.port}([A-Za-z_/]*)",stdout_value)
        if len(m) > 0:
            for gw in m:
                self.gatewaysArray.append(self.url + gw) 
                om.addGateway(self.url + gw)  
    
    #m = re.findall("findproperty <q>\[public\]::remObj(.*)",stdout_value)
    #if len(m) > 0:
    #   services.extend(m)
    
        services = []
        m = re.findall("destination id=\"([\w\d]*)\"",stdout_value)
        if len(m) > 0:
            services.extend(m)
            services = self.unique(services)
            self.servicesArray = services
            for s in services:
                om.addService(s) 
    
    
        m = re.findall("\[staticprotected]Object}::(.*), \d params",stdout_value)
        if len(m) > 0:
            methodnames = []
            unique_matches = self.unique(m)
            for name in unique_matches:
                if(name not in self.commonfunctions):
                    methodnames.append(name)
                    om.addMethod(name) 
            self.methodsArray = methodnames

